// 'use strict';
/**
 * Module dependencies.
 * @ignore
 */
var Request = require('./../../lib/request');
var Utils = require('./../../lib/utils');
var CacheProvider = require('./../cache-provider');

/**
 * Helper for expect and only
 * @api private
 * @ignore
 */
var _extend = function (type) {
    return function () {
        this._query[type] = this._query[type] || {};
        switch (arguments.length) {
            case 1:
                if (Array.isArray(arguments[0]) || typeof arguments[0] === "string") {
                    var query = this._query[type]['BASE'] || [];
                    query = query.concat(arguments[0]);
                    this._query[type]['BASE'] = query;
                    return this;
                } else {
                    console.error("Kindly provide valid parameters");
                }
                break;
            case 2:
                if (typeof arguments[0] === "string" && (Array.isArray(arguments[1]) || typeof arguments[1] === "string")) {
                    var query = this._query[type][arguments[0]] || [];
                    query = query.concat(arguments[1]);
                    this._query[type][arguments[0]] = query;
                    return this;
                } else {
                    console.error("Kindly provide valid parameters");
                }
                break;
            default:
                console.error("Kindly provide valid parameters");
        }
    };
};

/**
 * @summary Creates an instance of `Entry`.
 * @description An initializer is responsible for creating Entry object.
 * @param {String} uid - uid of the entry
 * @example
 * var Entry = Contentstack.Stack().ContentType('example).Entry();
 * @returns {Entry}
 * @ignore
 */
function Entry() {
    this._query = {};
    return this;
}

/**
 * @method only
 * @description This method is use to show the selected fields of the entries in resultset.
 * @param {String} [key=BASE] - reference field in the entry/single field in entry
 * @param {Array} values - array of fields to be show in resultset
 * @example
 * <caption> .only with field uid </caption>
 * blogEntry.only('title')
 * @example
 * <caption> .only with field uid </caption>
 * blogEntry.only('BASE','title')
 * @example
 * <caption> .only with field uids(array) </caption>
 * blogEntry.only(['title','description'])
 * @example
 * <caption> .only with reference_field_uid and field uid </caption>
 * blogEntry.only('category','title')
 * @example
 * <caption> .only with reference_field_uid and field uids(array) </caption>
 * blogEntry.only('category', ['title', 'description'])
 * @returns {Entry}
 */
Entry.prototype.only = _extend('only');

/**
 * @method except
 * @description This method is use to hide the selected fields of the entries in resultset.
 * @param {String} [key=BASE] - reference field in the entry/single field in entry
 * @param {Array} values - array of fields to be show in resultset
 * @example
 * <caption> .except with field uid </caption>
 * blogEntry.except('title')
 * @example
 * <caption> .except with field uid </caption>
 * blogEntry.except('BASE','title')
 * @example
 * <caption> .except with field uids(array) </caption>
 * blogEntry.except(['title','description'])
 * @example
 * <caption> .except with reference_field_uid and field uid </caption>
 * blogEntry.except('category','title')
 * @example
 * <caption> .except with reference_field_uid and field uids(array) </caption>
 * blogEntry.except('category', ['title', 'description'])
 * @returns {Entry} */
Entry.prototype.except = _extend('except');

/**
 * @method includeReference
 * @description This method is use to include referenced entries from the other Contenttype.
 * @example
 * <caption> .includeReference with reference_field_uids as array </caption>
 * blogEntry.includeReference(['category', 'author'])
 * @example
 * <caption> .includeReference with reference_field_uids </caption>
 * blogEntry.includeReference('category', 'author')
 * @returns {Entry}
 */
Entry.prototype.includeReference = function (val) {
    if (Array.isArray(val) || typeof val === "string") {
        this._query['include'] = this._query['include'] || [];
        this._query['include'] = this._query['include'].concat(val);
        return this;
    } else {
        console.error("Argument should be a String or an Array.");
    }
};

/**
 * @method language
 * @description This method is used set language code, which language's data to be retrieve.
 * @param {String} language_code - language code. e.g. 'en-us', 'ja-jp', etc.
 * @example blogEntry.language('en-us')
 * @returns {Entry}
 */
Entry.prototype.language = function (language_code) {
    if (language_code && typeof language_code === 'string') {
        this._query['locale'] = language_code;
        return this;
    } else {
        console.error("Argument should be a String.");
    }
};

/**
 * @method addQuery
 * @description This method is used to add query to Entry object.
 * @param {String} key - key of the query
 * @param {String} value - value of the query
 * @example blogEntry.addQuery('include_schema',true)
 * @returns {Entry}
 */
Entry.prototype.addQuery = function (key, value) {
    if (key && value && typeof key === 'string') {
        this._query[key] = value;
        return this;
    } else {
        console.error("First argument should be a String.");
    }
};

/**
 * @method includeSchema
 * @description This method is used to include the schema of the current contenttype in result set along with the entry/entries.
 * @example blogEntry.includeSchema()
 * @returns {Entry}
 */
Entry.prototype.includeSchema = function () {
    this._query['include_schema'] = true;
    return this;
};

/**
 * @method includeOwner
 * @description This method is used to include the owner of the entry/entries in resultset.
 * @example blogEntry.includeOwner()
 * @returns {Entry}
 */
Entry.prototype.includeOwner = function () {
    this._query['include_owner'] = true;
    return this;
};

/**
 * @method spread
 * @description This method is used to spread the result to the success callback with arguments.
 * @example
 * blogEntry.spread(function (entry, schema, count) {
 *      // success function
 * },function (error) {
 *      // error function
 * })
 * @returns {Entry}
 */
Entry.prototype.spread = function () {
    this._spread = true;
    if(this && this.then) this.then.apply(this, arguments);
};

/**
 * @method toJSON
 * @description This method is used to convert the result in to plain javascript object.
 * @example
 * blogEntry
 *      .toJSON()
 *      .then(function (result) {
 *          var value = result.get(field_uid)
*       },function (error) {
 *          // error function
 *      })
 * @returns {Object}
 */
Entry.prototype.toJSON = function () {
    this.tojson = true;
    return this;
};

/**
 * @method fetch
 * @description fetch entry of requested content_type of defined query if present.
 * @example
 * blogEntry.fetch()
 */
Entry.prototype.fetch = function () {
    if (this.entry_uid) {
        this.requestParams = {
            method: 'POST',
            headers: this.headers,
            url: this.config.protocol + "://" + this.config.host + '/' + this.config.version + this.config.urls.content_types + this.content_type_uid + this.config.urls.entries + this.entry_uid,
            body: {
                _method: 'GET',
                query: this._query
            }
        };
        return this;
    } else {
        console.error("Kindly provide an entry uid. e.g. .Entry('bltsomething123')");
    }
};

/**
 * @method then
 * @description Execute query and get data from Built.io Contentstack
 * @example
 * blogEntry.then(function () {
 *      // sucess function
 * },function () {
 *      // error function
 * })
 */
Entry.prototype.then = function (success, fail) {
    var env_uid = this.environment_uid,
        params = this.requestParams;
    if (env_uid) {
        this._query.environment_uid = env_uid;
    } else {
        this._query.environment = this.environment;
    }

    var self = this;
    var continueFlag = false;
    var cachePolicy = (typeof self.queryCachePolicy !== 'undefined') ? self.queryCachePolicy : self.cachePolicy;
    var tojson = (typeof self.tojson !== 'undefined') ? self.tojson : false;
    var spread = (typeof self._spread !== 'undefined') ? self._spread : false;
    var isSingle = (self.entry_uid || self.singleEntry) ? true : false;
    var hashQuery = Utils.getHash(Utils.parseQueryFromParams(self, isSingle, tojson));
    var spreadResult;

    var getCacheCallback = function () {
        return function (err, entries) {
            try {
                if (err) throw err;
                if (!tojson) entries = Utils.resultWrapper(entries);
                spreadResult = Utils.spreadResult(entries);
                if(isSingle) {
                    success(spreadResult);
                } else if (spread) {
                    success.apply(null, spreadResult);
                } else {
                    success(spreadResult);
                }
            } catch (e) {
                fail(e);
            }
        }
    };

    var callback = function (continueFlag) {
        if(continueFlag) {
            Request(params, function (err, data) {
                try {
                    self.queryCachePolicy = undefined;
                    self.tojson = undefined;
                    self._spread = undefined;
                    self.entry_uid = undefined;
                    if (!err) {
                        var entries = {};
                        if (this.singleEntry) {
                            if (data.entries && data.entries.length) {
                                entries.entry = Utils.getData(data, this.headers.access_token).entries[0];
                                if(data.schema) entries.schema = data.schema;
                            } else {
                                if(cachePolicy === 2) {
                                    self.provider.get(hashQuery, getCacheCallback());
                                } else {
                                    fail({ error_code: 141, error_message: 'The requested entry doesn\'t exist.' });
                                }
                                return;
                            }
                            this.singleEntry = false;
                        } else {
                            entries = Utils.getData(data, this.headers.access_token);
                        }
                        if(cachePolicy !== -1) {
                            self.provider.set(hashQuery, entries, function (err) {
                                try {
                                    if (err) throw err;
                                    if(!tojson) entries = Utils.resultWrapper(entries);
                                    spreadResult = Utils.spreadResult(entries);
                                    if(isSingle) {
                                        success(spreadResult);
                                    } else if (spread) {
                                        success.apply(null, spreadResult);
                                    } else {
                                        success(spreadResult);
                                    }
                                } catch(e) {
                                    fail(e);
                                }
                            });
                        } else {
                            if(!tojson) entries = Utils.resultWrapper(entries);
                            spreadResult = Utils.spreadResult(entries);
                            if(isSingle) {
                                success(spreadResult);
                            } else if (spread) {
                                success.apply(null, spreadResult);
                            } else {
                                success(spreadResult);
                            }
                        }
                    } else {
                        if(cachePolicy === 2) {
                            self.provider.get(hashQuery, getCacheCallback());
                        } else {
                            fail(err);
                        }
                    }
                } catch (e) {
                    fail({
                        message: e.message
                    });
                }
            }.bind(self));
        }
    };

    switch (cachePolicy) {
        case 1:
        case 3:
            self.provider.get(hashQuery, function (err, _data) {
                try {
                    var flag = false;
                    if(err) {
                        flag = true;
                        fail(err);
                    } else {
                        if(cachePolicy === 3 || !_data) flag = true;
                        if(_data) {
                            if (!tojson) _data = Utils.resultWrapper(_data);
                            spreadResult = Utils.spreadResult(_data);
                            if(isSingle) {
                                success(spreadResult);
                            } else if (spread) {
                                success.apply(null, spreadResult);
                            } else {
                                success(spreadResult);
                            }
                        } else if(cachePolicy === 3) {
                            success(_data);
                        }
                    }
                    callback(flag);
                } catch(e) {
                    fail(e);
                }
            });
            break;
        case 2:
        case 0:
        case -1:
            callback(true);
    };
};

module.exports = Entry;